'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var t = _ref.types;


  var buildTemplate = (0, _babelTemplate2.default)('\n    SYSTEM_GLOBAL.registerDynamic(MODULE_ID, [DEPS], false, BODY);\n  ');

  var buildFactory = (0, _babelTemplate2.default)('\n    (function($__require, $__exports, $__module) {\n      MODULE_URI\n      BODY;\n    })\n  ');

  var buildFactoryTypeCheck = (0, _babelTemplate2.default)('\n    FACTORY_DECLARATION\n    if (typeof FACTORY_REFERENCE === TYPE) {\n      return FACTORY_CALL;\n    } else {\n      return FACTORY_REFERENCE;\n    }\n  ');

  var buildFactoryExpressionDeclaration = (0, _babelTemplate2.default)('\n    var $__factory = FACTORY_EXPRESSION;\n  ');

  var buildModuleURIBinding = (0, _babelTemplate2.default)('\n     $__module.uri = $__module.id;\n  ');

  return {
    visitor: {
      CallExpression: function CallExpression(path, _ref2) {
        var _ref2$opts = _ref2.opts;
        var opts = _ref2$opts === undefined ? {} : _ref2$opts;

        var callee = path.node.callee;
        var args = path.node.arguments;

        // Leave nested define untouched
        var currentPath = path;
        while (currentPath.parentPath) {
          currentPath = currentPath.parentPath;
          if (currentPath.node.callee && t.isIdentifier(currentPath.node.callee.object, { name: 'System' }) && t.isIdentifier(currentPath.node.callee.property, { name: 'registerDynamic' })) {
            return;
          }
        }

        // match define(function(require) {})
        // match define(['dep1'], function(dep1) {})
        // match define('moduleName', function(dep1) {})
        // match define('moduleName', ['dep1'], function(dep1) {})
        if (t.isIdentifier(callee, { name: 'define' }) && !path.scope.hasBinding('define') && args.length >= 1 && args.length <= 3) {
          var callParams;

          (function () {
            var _deps;

            var moduleName = null,
                deps = [],
                factoryArg = null,
                isExportsInDeps = false,
                isModuleInDepsOrInFactoryParam = false;

            if (args.length === 1) {
              // first param is factory object/function
              factoryArg = args[0];
              // parse factory params
              if (t.isFunctionExpression(factoryArg)) {
                factoryArg.params.forEach(function (factoryParam) {
                  if (t.isIdentifier(factoryParam) && factoryParam.name === 'module') {
                    isModuleInDepsOrInFactoryParam = true;
                  }
                });
              }
            } else if (args.length === 2) {
              // first param is either module name or dependency array
              if (t.isStringLiteral(args[0])) {
                moduleName = args[0];
              } else if (t.isArrayExpression(args[0])) {
                deps = args[0];
              }
              // second param is factory object/function
              factoryArg = args[1];
            } else {
              // first param is module name
              moduleName = args[0];
              // second param is dependency array
              deps = args[1];
              // third param is factory object/function
              factoryArg = args[2];
            }

            // Call params used for the define factories wrapped in IIFEs
            callParams = [];

            if (t.isArrayExpression(deps)) {
              // Test if 'exports' exists as dependency: define(['exports'], function(exports) {})
              isExportsInDeps = deps.elements.filter(function (param) {
                return param.value === 'exports';
              }).length > 0;
              isModuleInDepsOrInFactoryParam = deps.elements.filter(function (param) {
                return param.value === 'module';
              }).length > 0;

              deps.elements.forEach(function (param) {
                if (['require', 'module', 'exports'].indexOf(param.value) !== -1) {
                  // Add all special identifiers in it's correct order and bind the to the param identifiers of the System.registerDynamic factory.
                  callParams.push(t.identifier('$__' + param.value));
                } else {
                  callParams.push(t.callExpression(t.identifier('$__require'), [param]));
                }
              });

              // Remove all special identifiers from the dependency list which will be provided by the System.registerDynamic factory.
              deps = deps.elements.filter(function (param) {
                return ['require', 'module', 'exports'].indexOf(param.value) === -1;
              });
            }

            var newDeps = [];

            // Handle CommonJS-sytle factories
            if (deps.length === 0 && t.isFunctionExpression(factoryArg)) {

              var detectThisAssignments = {
                // Visitor to determine all assignments to `this` members in a cjs factory function. If any `exports` must be used as `thisBindingExpression`
                ThisExpression: function ThisExpression(path) {

                  var currentPath = path;

                  // Traverse upwards until an assignment expression is reached
                  while (!t.isAssignmentExpression(currentPath.node) && currentPath.parentPath) {
                    currentPath = currentPath.parentPath;
                  }

                  if (t.isAssignmentExpression(currentPath.node)) {
                    var leftAssignmentOperand = currentPath.node.left;
                    if (t.isMemberExpression(leftAssignmentOperand)) {
                      leftAssignmentOperand = leftAssignmentOperand.object;
                      // Get base operand of member expression
                      while (leftAssignmentOperand.object) {
                        leftAssignmentOperand = leftAssignmentOperand.object;
                      }
                    }

                    // If leftAssignmentOperand is a `this` expression we have to use `$__exports` as `thisBindingExpression`
                    if (leftAssignmentOperand === path.node) {
                      isExportsInDeps = true;
                    }
                  }
                }
              };

              // We need to traverse over the complete path, since we didn't get the path of defineFactore here
              // TODO: Find a way to provide path of the cjs factory!!!
              path.traverse(detectThisAssignments, {});

              // Iterate over each param of the define factory
              factoryArg.params.forEach(function (param, index) {
                switch (index) {
                  // require
                  case 0:
                    var filterAMDDeps = {
                      // Visitor to determine all required dependencies if no DEPS array is provided but a factory function with present require param.
                      CallExpression: function CallExpression(path) {
                        var callee = path.node.callee;
                        var args = path.node.arguments;

                        // Get function scope of defineFactory where param is used
                        if (path.scope.bindings[param.name] && path.scope.bindings[param.name].identifier === param && t.isIdentifier(callee, { name: param.name })) {
                          newDeps.push(args[0]);
                        }
                      }
                    };

                    // We need to traverse over the complete path, since we didn't get the path of defineFactore here
                    // TODO: Find a way to provide path of the cjs factory!!!
                    path.traverse(filterAMDDeps, {});

                    // Push $__require to call params if used as factory default param
                    callParams.push(t.identifier('$__require'));
                    break;
                  // exports
                  case 1:
                    // Push $__require to call params if used as factory default param
                    callParams.push(t.identifier('$__exports'));
                    // Boolean flag which indicates that ```exports``` is present as factory param
                    isExportsInDeps = true;
                    break;
                  // module
                  case 2:
                    // Push $__require to call params if used as factory default param
                    callParams.push(t.identifier('$__module'));
                    break;
                }
              });
            }

            var factoryReferenceIdentifier = void 0,
                factoryAsExpressionDeclaration = void 0;
            var thisBindingExpression = isExportsInDeps ? t.identifier('$__exports') : t.thisExpression();
            if (t.isFunctionExpression(factoryArg)) {
              // If factory is passed as function argument
              var call = t.memberExpression(t.parenthesizedExpression(factoryArg), t.identifier('call'));
              factoryArg = t.callExpression(call, [thisBindingExpression].concat(_toConsumableArray(callParams)));
            } else if (t.isIdentifier(factoryArg)) {
              // If factory is passed as identifier argument
              factoryReferenceIdentifier = t.identifier(factoryArg.name);
              var _call = t.memberExpression(factoryArg, t.identifier('call'));
              factoryArg = t.callExpression(_call, [thisBindingExpression].concat(_toConsumableArray(callParams)));
            } else if (!t.isObjectExpression(factoryArg)) {
              // If factory is passed as expression argument
              factoryReferenceIdentifier = t.identifier('$__factory');
              factoryAsExpressionDeclaration = buildFactoryExpressionDeclaration({
                FACTORY_EXPRESSION: t.parenthesizedExpression(factoryArg)
              });
              var _call2 = t.memberExpression(factoryReferenceIdentifier, t.identifier('call'));
              factoryArg = t.callExpression(_call2, [thisBindingExpression].concat(_toConsumableArray(callParams)));
            }

            var factoryTypeTestNeeded = false;
            if ((!moduleName || moduleName && deps.length === 0) && factoryReferenceIdentifier) {
              // Wraps the factory in a ```if (typeof factory === 'function') {}``` test only a factory reference is present
              factoryTypeTestNeeded = true;
              factoryArg = buildFactoryTypeCheck({
                FACTORY_REFERENCE: factoryReferenceIdentifier,
                FACTORY_CALL: factoryArg,
                TYPE: t.stringLiteral('function'),
                FACTORY_DECLARATION: factoryAsExpressionDeclaration || null
              });
            }

            var factory = buildFactory({
              MODULE_URI: isModuleInDepsOrInFactoryParam ? buildModuleURIBinding() : null,
              BODY: factoryTypeTestNeeded ? factoryArg : t.returnStatement(factoryArg)
            });

            // Concat with required depencies array which contains string literals if factory default params are used
            deps = (_deps = deps).concat.apply(_deps, newDeps);

            // Map dependencies
            if (typeof opts.map === 'function') {
              deps.forEach(function (e) {
                e.value = opts.map(e.value);
              });
            }

            var systemRegister = buildTemplate({
              SYSTEM_GLOBAL: t.identifier('System'),
              MODULE_ID: moduleName,
              DEPS: [].concat(_toConsumableArray(deps)),
              BODY: factory
            });

            path.replaceWith(systemRegister);
          })();
        }
      },
      MemberExpression: function MemberExpression(path, _ref3) {
        var opts = _ref3.opts;

        // Replace `define.amd` with `true` if it's used inside a logical expression.
        if (t.isIdentifier(path.node.object, { name: 'define' }) && t.isIdentifier(path.node.property, { name: 'amd' }) && !path.scope.hasBinding('define') && path.parentPath && t.isLogicalExpression(path.parentPath)) {
          path.replaceWith(t.booleanLiteral(true));
        }
      },
      Identifier: function Identifier(path, _ref4) {
        var opts = _ref4.opts;

        // Replace `typeof define` if it's used inside a unary expression.
        if (t.isIdentifier(path.node, { name: 'define' }) && !path.scope.hasBinding('define') && path.parentPath && t.isUnaryExpression(path.parentPath) && path.parentPath.node.operator === 'typeof') {
          path.parentPath.replaceWith(t.stringLiteral('function'));
        }
      }
    }
  };
};

var _babelTemplate = require('babel-template');

var _babelTemplate2 = _interopRequireDefault(_babelTemplate);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }